//
//----------------------------------------------
// Programm für FREMO Ssp an DIN-Stecker - v8.1
// verändert von Bodo Mertins
//----------------------------------------------

// Servo bei Installation mit Hebel in Horizontalstellung und Servo mit 90° bzw. Mittelstellung (Servotester) montieren!

#include <Servo.h>
#include <EEPROM.h>


// Fogende Konstanten NICHT ändern:
#define def_MelderBeiFreigabeOderEntnommen         0
#define def_MelderNurBeiFreigabe                   1
#define def_KeinMelderBeiEntnommen                 2
#define def_MelderBeiFreigabeUNDSchluesselSicher   3



///////////////////////////////////////////////////////////////////////////////////
// EINSTELLUNGEN:

// NUR EINE der vier vorherigen Definitionen eintragen:
#define Melderverhalten        def_MelderBeiFreigabeUNDSchluesselSicher
//#define Melderverhalten        def_KeinMelderBeiEntnommen
//#define Melderverhalten        def_MelderBeiFreigabeOderEntnommen
//#define Melderverhalten        def_MelderNurBeiFreigabe

#define Schutzzeit_ms          250

#define JPRepeat_ms            500

#define REMVerzoegerung_ms    1200

#define Fall2aMoeglich         true

// Ende Einstellungen!
///////////////////////////////////////////////////////////////////////////////////

#define ServoInitPos_Open     75
#define ServoClosePosOffset   50
// Servoposition "Sl. entnehmbar" -   2040: (90) | 0212: (75)  / Mögliches Intervall: (65...100) deg
// Servoposition "Sl. verschlossen" - 2040: (140)| 0212: (125) / Betrag ab "Pos_Open": 50 deg

#define ServoEEPROMAdr        16

const int SchluesselKontakt  = 3;   // Mikroschalter zum Überprüfung der Schlüssellage
const int TasterSsp          = 4;

const int LEDSsp             = 5;

const int PlusJumperEingang  = 6;
const int MinusJumperEingang = 7;

const int Relais             = 8;   // Relaiskontakt für DIN 3/5 --> Bestätigung geschl. Ssp an Stw
const int FreigabeEingang    = 9;   // Kontakt/Schalter DIN 1/4 <-- Freigabe Ssp von Stw


const int Servo_Pin          = 2;
Servo SperrServo;



#define bmslipEnd         192
#define bmslipEsc         219
#define bmslipEndErsatz   220
#define bmslipEscErsatz   221


int ZeitSperre = Schutzzeit_ms;
bool LastSchluesselSicher = false;
bool GespeicherteFreigabe = false;
bool HilfMelderErlaubt    = true;
int JPZaehlerP=0;
int JPZaehlerM=0;
int REMZaehler=0;
bool LastPlusJP = false;
bool LastMinusJP = false;
unsigned char ServoOpenPos = ServoInitPos_Open;

#define ComInMax 8
char ComInBuf[ComInMax];
char ComInAnz = 0;
char ComInEsc = 0;



void IncServo(void)
{
  if (ServoOpenPos<100)
  {
    ServoOpenPos++;
    EEPROM.update(ServoEEPROMAdr,ServoOpenPos);
  }
}

void DecServo(void)
{
  if (ServoOpenPos>65)
  {
    ServoOpenPos--;
    EEPROM.update(ServoEEPROMAdr,ServoOpenPos);
  }
}

void SetServo(int w)
{
  ServoOpenPos=w; // 65..100
  if ((ServoOpenPos<65) || (ServoOpenPos>100))
    ServoOpenPos=ServoInitPos_Open;
  EEPROM.update(ServoEEPROMAdr,ServoOpenPos);
}



void setup() {
  int i;
  
  pinMode(TasterSsp, INPUT_PULLUP);          // LOW == gedrückt
  pinMode(SchluesselKontakt, INPUT_PULLUP);  // NEU v7.5: HIGH == gedrückt
  pinMode(FreigabeEingang, INPUT_PULLUP);    // LOW == gedrückt
  pinMode(PlusJumperEingang, INPUT_PULLUP);  // LOW == gedrückt
  pinMode(MinusJumperEingang, INPUT_PULLUP); // LOW == gedrückt
  
  pinMode(LEDSsp, OUTPUT);
  digitalWrite(LEDSsp, LOW);  // LOW == aus

  //pinMode(LEDStw, OUTPUT);
  //digitalWrite(LEDStw, LOW);  // LOW == aus

  pinMode(Relais, OUTPUT);
  digitalWrite(Relais, LOW);  // LOW == aus = Kontakt offen = "Schlüssel frei"

  ServoOpenPos=EEPROM.read(ServoEEPROMAdr);
  SetServo(ServoOpenPos); // Wegen Grenzen/Initialisierung!

  SperrServo.attach(Servo_Pin);
  SperrServo.write(ServoOpenPos);

  ComInAnz=0;
  ComInEsc=0;
  Serial.begin(9600);
  while (!Serial);
}

void AddRaw(unsigned char b)
{
  Serial.write(b);
}

void Add(unsigned char b)
{
  switch (b)
  {
  case bmslipEnd:AddRaw(bmslipEsc);AddRaw(bmslipEndErsatz);break;
  case bmslipEsc:AddRaw(bmslipEsc);AddRaw(bmslipEscErsatz);break;
  default:AddRaw(b);
  } // of switch
}

void SendePaket(unsigned char *s,unsigned char len)
{
  int i,ps;
  unsigned char b;

  AddRaw(bmslipEnd);
  ps=0;
  for (i=0;i<len;i++)
  {
    b=s[i];
    ps+=b;
    Add(b);
  }
  Add(ps&255);
  AddRaw(bmslipEnd);
}



void loop() {
  bool FreigabeVomStw;
  bool TasteGedrueckt;
  bool SchluesselSicher;
  bool ServoSollOffenSein;
  bool MelderFlag;
  bool GesamtFreigabe;
  bool AktPlusJP;
  bool AktMinusJP;
  unsigned char b;
  int i,ps;
  unsigned char Antwort[8];
  
  delay(1); // Damit dauert ein Loop gut 1 ms!
  
  if (ZeitSperre>0) {
    ZeitSperre--;
  }



  FreigabeVomStw   = (digitalRead(FreigabeEingang) == LOW);
  TasteGedrueckt   = (digitalRead(TasterSsp) == LOW);
  SchluesselSicher = (digitalRead(SchluesselKontakt) == LOW); // NEU v7.5 == LOW


  if (digitalRead(PlusJumperEingang) == LOW)
  {
    if (JPZaehlerP<JPRepeat_ms)
      JPZaehlerP++;
  }
  else
    JPZaehlerP=0;
  
  if (digitalRead(MinusJumperEingang) == LOW)
  {
    if (JPZaehlerM<JPRepeat_ms)
      JPZaehlerM++;
  }
  else
    JPZaehlerM=0;

  AktPlusJP        = (JPZaehlerP == JPRepeat_ms) && (JPZaehlerM < JPRepeat_ms);
  AktMinusJP       = (JPZaehlerM == JPRepeat_ms) && (JPZaehlerP < JPRepeat_ms);

  if (AktPlusJP && (!LastPlusJP))
  {
    IncServo();
    JPZaehlerP=0;
    JPZaehlerM=0;
  }
  if (AktMinusJP && (!LastMinusJP))
  {
    DecServo();
    JPZaehlerP=0;
    JPZaehlerM=0;
  }

  LastPlusJP=AktPlusJP;
  LastMinusJP=AktMinusJP;



  while (Serial.available()>0)
  {
    b=Serial.read();
    switch (b)
    {
    case bmslipEnd:{
                     if (ComInAnz>1)
                     { // Auswertung
                       ps=0;
                       for (i=0;i<ComInAnz-1;i++) // ohne Prüfsumme!
                         ps+=ComInBuf[i];

                       if (ComInBuf[ComInAnz-1]==ps)
                       {
                         switch (ComInBuf[0])
                         {
                         case 40:SetServo(ServoInitPos_Open);break; // Init 0212
                         case 41:SetServo(90);break;                // Init 2040
                         case 42:{
                                   Antwort[0]=42;
                                   Antwort[1]=ServoOpenPos;
                                   SendePaket(&Antwort[0],2);
                                   break;
                                 }
                         case 43:IncServo();break;
                         case 44:DecServo();break;
                         } // switch
                       }
                     }
                     ComInAnz=0;
                     ComInEsc=0;
                     break;
                   }
    case bmslipEsc:ComInEsc=1;break;
    default:{
              if (ComInEsc)
              {
                switch (b)
                {
                case bmslipEndErsatz:b=bmslipEnd;break;
                case bmslipEscErsatz:b=bmslipEsc;break;
                } // switch
              }
              ComInEsc=0;
              if (ComInAnz<ComInMax)
                ComInBuf[ComInAnz++]=b;
            }
    } // switch

    //Serial.write(b);
  }




  if (Fall2aMoeglich)
  {
    if (FreigabeVomStw && TasteGedrueckt)
      GespeicherteFreigabe = true;
    else if (!TasteGedrueckt)
      GespeicherteFreigabe = false;
  }
  else // Ohne Fall 2a:
    GespeicherteFreigabe = (FreigabeVomStw && TasteGedrueckt);
  GesamtFreigabe = (GespeicherteFreigabe || FreigabeVomStw); // Nur zur Übersichtlichkeit!

  
  ServoSollOffenSein = GespeicherteFreigabe || (!SchluesselSicher); // Öffnen ODER entnommenen Schlüssel offen halten
  if (ServoSollOffenSein)
    SperrServo.write(ServoOpenPos); // Geht immer!
  else
  { // Servo soll locked sein:
    if (!LastSchluesselSicher) // EINMAL...
      ZeitSperre = Schutzzeit_ms;
    if (ZeitSperre == 0)
      SperrServo.write(ServoOpenPos+ServoClosePosOffset); 
  }
  LastSchluesselSicher = SchluesselSicher;


  if (SchluesselSicher && (!ServoSollOffenSein) && (ZeitSperre == 0))
    digitalWrite(Relais, HIGH); // "Schlüssel sicher"
  else
    digitalWrite(Relais, LOW); // "Schlüssel frei/entnommen"


  // Die folgende Konstruktion verhindert das Aufblitzen des Melders beim Wiedereinschließen ;-)
  if (!SchluesselSicher)
    HilfMelderErlaubt = false;
  else if (SchluesselSicher && (!GesamtFreigabe))
    HilfMelderErlaubt = true;
  
  switch(Melderverhalten)
  {
  case def_MelderBeiFreigabeUNDSchluesselSicher:MelderFlag=GesamtFreigabe && SchluesselSicher;break;
  case def_KeinMelderBeiEntnommen:              MelderFlag=GesamtFreigabe && HilfMelderErlaubt;break;
  case def_MelderBeiFreigabeOderEntnommen:      MelderFlag=GesamtFreigabe || (!SchluesselSicher);break;
  case def_MelderNurBeiFreigabe:                MelderFlag=GesamtFreigabe;break;
  default:MelderFlag=false; // Sollte nicht auftreten!
  }
  
  if (MelderFlag)
  {
    if (REMZaehler<REMVerzoegerung_ms)
      REMZaehler++;
  }
  else
    REMZaehler=0;
  
  if (REMZaehler>=REMVerzoegerung_ms)
    digitalWrite(LEDSsp, HIGH); // Ein 
  else
    digitalWrite(LEDSsp, LOW); // Aus
}
